"""
This module contains tools for rasterizing vector data.
"""
from typing import Optional, Union

import geopandas
import numpy
import odc.geo.geobox
import pandas
import rasterio
import rasterio.features
from numpy.typing import NDArray
from packaging import version
from rasterio.enums import MergeAlg
from scipy.interpolate import Rbf, griddata

_INT8_SUPPORTED = version.parse(rasterio.__gdal_version__) >= version.parse(
    "3.7.0"
) and version.parse(rasterio.__version__) >= version.parse("1.3.7")


def _is_numeric(data_values: NDArray) -> bool:
    """
    Check if array data type is numeric.
    """
    return numpy.issubdtype(data_values.dtype.type, numpy.number)


def _remove_missing_data(
    data_values: NDArray,
    geometry_array: geopandas.GeoSeries,
) -> tuple[NDArray, geopandas.GeoSeries]:
    """
    Missing data causes issues with interpolation of point data
    https://github.com/corteva/geocube/issues/9

    This filters the data so those issues don't cause problems.
    """
    not_missing_data = ~pandas.isnull(data_values)
    geometry_array = geometry_array[not_missing_data]
    data_values = data_values[not_missing_data]
    return data_values, geometry_array


def _minimize_dtype(dtype: numpy.dtype, fill: float) -> numpy.dtype:
    """
    If int64, convert to float64:
    https://github.com/OSGeo/gdal/issues/3325

    Attempt to convert to float32 if fill is NaN and dtype is integer.
    """
    if numpy.issubdtype(dtype, numpy.integer):
        if not _INT8_SUPPORTED and dtype.name == "int8":
            # GDAL<3.7/rasterio<1.3.7 doesn't support int8
            dtype = numpy.dtype("int16")
        if numpy.isnan(fill):
            dtype = (
                numpy.dtype("float64") if dtype.itemsize > 2 else numpy.dtype("float32")  # type: ignore
            )
    elif not numpy.issubdtype(dtype, numpy.floating):
        # default to float64 for non-integer/float types
        dtype = numpy.dtype("float64")
    return dtype


def rasterize_image(
    geometry_array: geopandas.GeoSeries,
    data_values: Union[NDArray, pandas.arrays.IntegerArray],
    geobox: odc.geo.geobox.GeoBox,
    fill: float,
    merge_alg: MergeAlg = MergeAlg.replace,
    filter_nan: bool = False,
    all_touched: bool = False,
    **ignored_kwargs,
) -> Optional[NDArray]:
    """
    Rasterize a list of shapes+values for a given GeoBox.

    Parameters
    -----------
    geometry_array: geopandas.GeoSeries
        A geometry array of points.
    data_values: Union[NDArray, pandas.arrays.IntegerArray]
        Data values associated with the list of geojson shapes
    geobox: :obj:`odc.geo.geobox.GeoBox`
        Transform of the resulting image.
    fill: float
        The value to fill in the grid with for nodata.
    merge_alg: `rasterio.enums.MergeAlg`, optional
        The algorithm for merging values into one cell. Default is `MergeAlg.replace`.
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    all_touched: bool, optional
        Passed to rasterio.features.rasterize. If True, all pixels touched by
        geometries will be burned in. If false, only pixels whose center is
        within the polygon or that are selected by Bresenham’s line algorithm
        will be burned in.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :obj:`numpy.ndarray` or None
        The vector data in the rasterized format.

    """
    if not _is_numeric(data_values):
        # only numbers can be rasterized
        return None

    if isinstance(data_values, pandas.arrays.IntegerArray):
        data_values = data_values.to_numpy(
            dtype=_minimize_dtype(data_values.dtype.numpy_dtype, fill),
            na_value=fill,
        )

    if filter_nan:
        data_values, geometry_array = _remove_missing_data(data_values, geometry_array)

    image = rasterio.features.rasterize(
        zip(geometry_array.values, data_values),
        out_shape=(geobox.height, geobox.width),
        transform=geobox.affine,
        fill=fill,
        all_touched=all_touched,
        merge_alg=merge_alg,
        dtype=_minimize_dtype(data_values.dtype, fill),
    )
    return image


def rasterize_points_griddata(
    geometry_array: geopandas.GeoSeries,
    data_values: NDArray,
    grid_coords: dict[str, NDArray],
    fill: float,
    method: str = "nearest",
    rescale: bool = False,
    filter_nan: bool = False,
    **ignored_kwargs,
) -> Optional[NDArray]:
    """
    This method uses scipy.interpolate.griddata to interpolate point data
    to a grid.

    Parameters
    ----------
    geometry_array: geopandas.GeoSeries
        A geometry array of points.
    data_values: list
        Data values associated with the list of geojson shapes
    grid_coords: dict
        Output from `rioxarray.rioxarray.affine_to_coords`
    fill: float
        The value to fill in the grid with for nodata.
    method: {'linear', 'nearest', 'cubic'}, optional
        The method to use for interpolation in `scipy.interpolate.griddata`.
    rescale: bool, optional
        Rescale points to unit cube before performing interpolation. Default is false.
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :class:`numpy.ndarray`: An interpolated :class:`numpy.ndarray`.

    """
    if not _is_numeric(data_values):
        # only numbers can be rasterized
        return None

    if filter_nan:
        data_values, geometry_array = _remove_missing_data(data_values, geometry_array)

    return griddata(
        points=(geometry_array.x, geometry_array.y),
        values=data_values,
        xi=tuple(numpy.meshgrid(grid_coords["x"], grid_coords["y"])),
        method=method,
        fill_value=fill,
        rescale=rescale,
    )


def rasterize_points_radial(
    geometry_array: geopandas.GeoSeries,
    data_values: NDArray,
    grid_coords: dict[str, NDArray],
    method: str = "linear",
    filter_nan=False,
    **ignored_kwargs,
) -> Optional[NDArray]:
    """
    This method uses scipy.interpolate.Rbf to interpolate point data
    to a grid.

    Parameters
    ----------
    geometry_array: geopandas.GeoSeries
        A geometry array of points.
    data_values: list
        Data values associated with the list of geojson shapes
    grid_coords: dict
        Output from `rioxarray.rioxarray.affine_to_coords`
    method: str, optional
        The function to use for interpolation in `scipy.interpolate.Rbf`.
        {'multiquadric', 'inverse', 'gaussian', 'linear',
        'cubic', 'quintic', 'thin_plate'}
    filter_nan: bool, optional
        If True, will remove nodata values from the data before rasterization.
        Default is False.
    **ignored_kwargs:
        These are there to be flexible with additional rasterization methods and
        will be ignored.

    Returns
    -------
    :class:`numpy.ndarray`: An interpolated :class:`numpy.ndarray`.

    """
    if not _is_numeric(data_values):
        # only numbers can be rasterized
        return None

    if filter_nan:
        data_values, geometry_array = _remove_missing_data(data_values, geometry_array)

    interp = Rbf(geometry_array.x, geometry_array.y, data_values, function=method)
    return interp(*numpy.meshgrid(grid_coords["x"], grid_coords["y"]))



"""
GeoCube client core functionality
"""
import os
from collections.abc import Iterable
from typing import Any, Callable, Literal, Optional, Union

import geopandas
import numpy
import shapely
import xarray
from odc.geo.geom import Geometry

from geocube.geo_utils.geobox import GeoBoxMaker
from geocube.vector_to_cube import VectorToCube


def make_geocube(
    vector_data: Union[str, os.PathLike, geopandas.GeoDataFrame],
    measurements: Optional[list[str]] = None,
    datetime_measurements: Optional[list[str]] = None,
    output_crs: Any = None,
    resolution: Optional[Union[float, Iterable[float]]] = None,
    align: Optional[tuple[float, float]] = None,
    geom: Optional[
        Union[str, dict, shapely.geometry.base.BaseGeometry, Geometry]
    ] = None,
    like: Optional[Union[xarray.Dataset, xarray.DataArray]] = None,
    fill: float = numpy.nan,
    group_by: Optional[str] = None,
    interpolate_na_method: Optional[Literal["linear", "nearest", "cubic"]] = None,
    categorical_enums: Optional[dict[str, list]] = None,
    rasterize_function: Optional[Callable[..., Optional[numpy.typing.NDArray]]] = None,
) -> xarray.Dataset:
    """
    Rasterize vector data into an ``xarray`` object.  Each attribute will be a data
    variable in the :class:`xarray.Dataset`.

    Parameters
    ----------
    vector_data: str, path-like object or :obj:`geopandas.GeoDataFrame`
        A file path to an OGR supported source or GeoDataFrame containing
        the vector data.
    measurements: list(str), optional
        Attributes name or list of names to be included. If a list is specified,
        the attributes will be returned in the order requested.
        By default all available attributes are included.
    datetime_measurements: list(str), optional
        Attributes that are temporal in nature and should be converted to the datetime
        format. These are only included if listed in 'measurements'.
    output_crs: Any, optional
        The CRS of the returned data. Can be anything accepted by :obj:`pyproj.CRS`.
        If no CRS is supplied, the CRS of the stored data is used.
    resolution: Union[float, Iterable[float]], optional
        A tuple of the spatial resolution of the returned data (Y, X).
        This includes the direction (as indicated by a positive or negative number).
        Typically when using most CRSs, the first number would be negative.
    align: tuple[float, float], optional
        Load data such that point 'align' lies on the pixel boundary.
        Units are in the co-ordinate space of the output CRS.
        Default is (0,0)
    geom: Union[str, dict, shapely.geometry.base.BaseGeometry, odc.geo.geom.Geometry], optional
        A GeoJSON string for the bounding box of the data used to construct the
        grid. It defaults to EPSG:4326 if a CRS is not provided.
        Example of adding CRS::

            {"type": "Polygon", "crs": {"properties": {"name": "EPSG:3857"}}}

    like: :obj:`xarray.Dataset`, optional
        Uses the output of a previous ``load()`` to form the basis of a request
        for another product.
        E.g.::

            gcds = make_geocube(vector_data='my_vector.geopackage', like=other_gcds)

    fill: float, optional
        The value to fill in the grid with for nodata. Default is NaN.
    group_by: str, optional
        When specified, perform basic combining/reducing of the data on this column.
    interpolate_na_method:  {'linear', 'nearest', 'cubic'}, optional
        This is the method for interpolation to use to fill in the nodata with
        :meth:`scipy.interpolate.griddata`.
    categorical_enums: dict, optional
        A dictionary of all categories for the table columns containing categorical
        data. The categories will be made unique and sorted if they are not already.
        E.g. {'column_name': ['a', 'b'], 'other_column': ['c', 'd']}
    rasterize_function: function, optional
        Function to rasterize geometries. Other options are available in
        `geocube.rasterize`. Default is :func:`geocube.rasterize.rasterize_image`.

    Returns
    --------
    :obj:`xarray.Dataset`:
        Requested data in a :obj:`xarray.Dataset`.

    """
    geobox_maker = GeoBoxMaker(output_crs, resolution, align, geom, like)

    return VectorToCube(
        vector_data=vector_data,
        geobox_maker=geobox_maker,
        fill=fill,
        categorical_enums=categorical_enums,
        measurements=measurements,
        group_by=group_by,
        datetime_measurements=datetime_measurements,
    ).make_geocube(
        interpolate_na_method=interpolate_na_method,
        rasterize_function=rasterize_function,
    )



"""
Utility methods to print system info for debugging

adapted from :func:`sklearn.utils._show_versions`
which was adapted from :func:`pandas.show_versions`
"""
# pylint: disable=import-outside-toplevel
import importlib.metadata
import platform
import sys


def _get_sys_info():
    """System information
    Return
    ------
    sys_info : dict
        system and Python version information
    """
    blob = [
        ("python", sys.version.replace("\n", " ")),
        ("executable", sys.executable),
        ("machine", platform.platform()),
    ]

    return dict(blob)


def _get_gdal_info():
    """Get the GDAL dependency information.

    Returns
    -------
    proj_info: dict
        system GDAL information
    """
    import fiona
    import rasterio

    blob = [
        ("fiona", importlib.metadata.version("fiona")),
        ("GDAL[fiona]", fiona.__gdal_version__),
        ("rasterio", importlib.metadata.version("rasterio")),
        ("GDAL[rasterio]", rasterio.__gdal_version__),
    ]

    return dict(blob)


def _get_deps_info():
    """Overview of the installed version of dependencies
    Returns
    -------
    deps_info: dict
        version information on relevant Python libraries
    """
    deps = [
        "appdirs",
        "click",
        "geopandas",
        "odc_geo",
        "rioxarray",
        "pyproj",
        "xarray",
    ]

    def get_version(module):
        try:
            return importlib.metadata.version(module)
        except importlib.metadata.PackageNotFoundError:
            return None

    return {dep: get_version(dep) for dep in deps}


def _print_info_dict(info_dict):
    """Print the information dictionary"""
    for key, stat in info_dict.items():
        print(f"{key:>14}: {stat}")


def show_versions():
    """
    .. versionadded:: 0.0.12

    Print useful debugging information

    Example
    -------
    > python -c "import geocube; geocube.show_versions()"

    """
    print(f"geocube v{importlib.metadata.version('geocube')}\n")
    print("GDAL deps:")
    _print_info_dict(_get_gdal_info())
    print("\nPython deps:")
    _print_info_dict(_get_deps_info())
    print("\nSystem:")
    _print_info_dict(_get_sys_info())



"""
This module is for GIS related utility functions.
"""
import json
import os
from collections.abc import Iterable
from typing import Any, Optional, Union

import fiona.errors
import geopandas
import rioxarray  # noqa: F401 pylint: disable=unused-import
import shapely.geometry.base
import xarray
from odc.geo import resyx_, wh_
from odc.geo.crs import CRS
from odc.geo.geobox import GeoBox
from odc.geo.geom import Geometry
from shapely.geometry import box, mapping

from geocube.exceptions import VectorDataError
from geocube.logger import get_logger


def geobox_from_rio(xds: Union[xarray.Dataset, xarray.DataArray]) -> GeoBox:
    """This function retrieves the geobox using rioxarray extension.

    Parameters
    ----------
    xds: :obj:`xarray.DataArray` or :obj:`xarray.Dataset`
        The xarray dataset to get the geobox from.

    Returns
    -------
    :obj:`odc.geo.geobox.GeoBox`

    """
    height, width = xds.rio.shape
    try:
        transform = xds.rio.transform()
    except AttributeError:
        transform = xds[xds.rio.vars[0]].rio.transform()
    return GeoBox(
        shape=wh_(width, height),
        affine=transform,
        crs=CRS(xds.rio.crs.to_wkt()),
    )


def load_vector_data(
    vector_data: Union[str, os.PathLike, geopandas.GeoDataFrame],
    measurements: Optional[list[str]] = None,
) -> geopandas.GeoDataFrame:
    """
    Parameters
    ----------
    vector_data: str, path-like object or :obj:`geopandas.GeoDataFrame`
        A file path to an OGR supported source or GeoDataFrame containing
        the vector data.
    measurements: list[str], optional
        Attributes name or list of names to be included. If a list is specified,
        the measurements will be returned in the order requested.
        By default all available measurements are included.

    Returns
    -------
    :obj:`geopandas.GeoDataFrame` containing the vector data.

    """
    logger = get_logger()

    if isinstance(vector_data, (str, os.PathLike)):
        try:
            vector_data = geopandas.read_file(vector_data, include_fields=measurements)
        except fiona.errors.DriverError as error:
            if "ignore_fields" not in str(error):
                raise
            vector_data = geopandas.read_file(vector_data)

    elif not isinstance(vector_data, geopandas.GeoDataFrame):
        vector_data = geopandas.GeoDataFrame(vector_data)

    if vector_data.empty:
        raise VectorDataError("Empty GeoDataFrame.")

    if measurements is not None:
        vector_data = vector_data[measurements + [vector_data.geometry.name]]

    try:
        vector_data.geometry
    except AttributeError as error:
        raise VectorDataError(
            "'geometry' column missing. Columns in file: "
            f"{vector_data.columns.values.tolist()}"
        ) from error

    # make sure projection is set
    if not vector_data.crs:
        vector_data.crs = "EPSG:4326"
        logger.warning(
            "Projection not defined in `vector_data`."
            " Setting to geographic (EPSG:4326)."
        )
    return vector_data


class GeoBoxMaker:
    """
    This class is meant for delayed GeoBox making. Stores partial information until
    all information needed is obtained.
    """

    def __init__(
        self,
        output_crs: Any,
        resolution: Optional[Union[float, Iterable[float]]],
        align: Optional[tuple[float, float]],
        geom: Optional[Union[str, dict, shapely.geometry.base.BaseGeometry, Geometry]],
        like: Optional[Union[xarray.Dataset, xarray.DataArray]],
    ) -> None:
        """Get the geobox to use for the grid.

        Parameters
        ----------
        output_crs: Any, optional
            The CRS of the returned data.  If no CRS is supplied, the CRS of
             the stored data is used.
        resolution: Union[float, Iterable[float]], optional
            A tuple of the spatial resolution of the returned data.
            This includes the direction (as indicated by a positive or negative number).
            Typically when using most CRSs, the first number would be negative.
        align: tuple[float, float], optional
            Load data such that point 'align' lies on the pixel boundary.
            Units are in the co-ordinate space of the output CRS.
            Default is (0,0)
        geom: Union[str, dict, shapely.geometry.base.BaseGeometry, odc.geo.geom.Geometry], optional
            A GeoJSON string for the bounding box of the data.
        like: :obj:`xarray.Dataset` | :obj:`xarray.DataArray`, optional
            Uses the output of a previous ``load()`` to form the basis of a request for
            another product.

        """
        self.output_crs = output_crs
        self.resolution = (
            resolution if not isinstance(resolution, Iterable) else resyx_(*resolution)
        )
        self.align = None if align is None else resyx_(*align)

        if isinstance(geom, str):
            geom = json.loads(geom)
        self.geom = geom
        self.like = like

    def from_vector(
        self, vector_data: Union[str, os.PathLike, geopandas.GeoDataFrame]
    ) -> GeoBox:
        """Get the geobox to use for the grid.

        Parameters
        ----------
        vector_data: str, path-like object or :obj:`geopandas.GeoDataFrame`
            A file path to an OGR supported source or GeoDataFrame
            containing the vector data.

        Returns
        -------
        :obj:`odc.geo.geobox.GeoBox`
            The geobox for the grid to be generated from the vector data.

        """
        vector_data = load_vector_data(vector_data)

        if self.like is not None:
            assert (
                self.output_crs is None
            ), "'like' and 'output_crs' are not supported together"
            assert (
                self.resolution is None
            ), "'like' and 'resolution' are not supported together"
            assert self.align is None, "'like' and 'align' are not supported together"
            try:
                geobox = self.like.geobox
            except (AttributeError, ValueError):
                geobox = geobox_from_rio(self.like)
            return geobox

        if self.resolution is None:
            raise RuntimeError("Must specify 'resolution' if 'like' not specified.")

        if self.output_crs:
            crs = CRS(self.output_crs)
        else:
            crs = CRS(vector_data.crs)

        if self.geom is None and self.output_crs:
            geopoly = Geometry(
                mapping(box(*vector_data.to_crs(crs).total_bounds)),
                crs=crs,
            )
        elif self.geom is None:
            geopoly = Geometry(mapping(box(*vector_data.total_bounds)), crs=crs)

        else:
            crs_input = "EPSG:4326"
            if isinstance(self.geom, dict) and "crs" in self.geom:
                crs_input = self.geom["crs"]["properties"]["name"]
            geom_crs = CRS(crs_input)
            geopoly = Geometry(self.geom, crs=geom_crs)
        return GeoBox.from_geopolygon(geopoly, self.resolution, crs, self.align)



"""
geocube core conversion functionality
"""
import os
from typing import Any, Callable, Literal, Optional, Union

import geopandas
import numpy
import pandas
import xarray
from numpy.typing import NDArray
from rioxarray.rioxarray import DEFAULT_GRID_MAP, affine_to_coords

from geocube.geo_utils.geobox import GeoBoxMaker, load_vector_data
from geocube.logger import get_logger
from geocube.rasterize import rasterize_image


def _format_series_data(data_series: geopandas.GeoSeries) -> geopandas.GeoSeries:
    """
    The purpose of this function is to convert the series data into a rasterizeable
    format if possible.

    Parameters
    ----------
    data_series: :obj:`geopandas.GeoSeries`
        The series to be converted.

    Returns
    -------
    :obj:`geopandas.GeoSeries`: The series that was converted if possible.

    """
    if "datetime" in str(data_series.dtype):
        data_series = pandas.to_numeric(data_series).astype(numpy.float64)
        get_logger().warning(
            f"The series '{data_series.name}' was converted from a date to a number to "
            "rasterize the data. To load the data back in as a date, "
            "use 'pandas.to_datetime()'."
        )
    elif str(data_series.dtype) == "category":
        data_series = data_series.cat.codes
    return data_series


class VectorToCube:
    # pylint: disable=too-many-instance-attributes
    """
    Tool that facilitates converting vector data to raster data into
    an :obj:`xarray.DataFrame`.
    """

    def __init__(
        self,
        vector_data: Union[str, os.PathLike, geopandas.GeoDataFrame],
        geobox_maker: GeoBoxMaker,
        fill: float,
        categorical_enums: Optional[dict[str, list]],
        measurements: Optional[list[str]] = None,
        datetime_measurements: Optional[list[str]] = None,
        group_by: Optional[str] = None,
    ):
        """
        Initialize the GeoCube class.

        Parameters
        ----------
        vector_data: str, path-like object or :obj:`geopandas.GeoDataFrame`
            A file path to an OGR supported source or GeoDataFrame containing
            the vector data.
        geobox_maker: :obj:`geocube.geo_utils.geobox.GeoBoxMaker`
            The geobox for the grid to be generated from the vector data.
        fill: float, optional
            The value to fill in the grid with for nodata. Default is NaN.
        categorical_enums: dict, optional
            A dictionary of all categories for the table columns containing
            categorical data. The categories will be made unique and sorted
            if they are not already.
            E.g. {'column_name': ['a', 'b'], 'other_column': ['c', 'd']}
        measurements: list[str], optional
            Attributes name or list of names to be included. If a list is specified,
            the measurements will be returned in the order requested.
            By default all available measurements are included.
        datetime_measurements: list[str], optional
            Attributes that are temporal in nature and should be converted to the
            datetime format. These are only included if listed in 'measurements'.
        group_by: str, optional
            When specified, perform basic combining/reducing of the data on this column.
        """
        self._fill = fill if fill is not None else numpy.nan
        self._group_by = group_by
        self._rasterize_function: Callable[..., Optional[NDArray]] = rasterize_image

        load_measurements = measurements
        if (
            load_measurements is not None
            and self._group_by is not None
            and self._group_by not in load_measurements
        ):
            load_measurements.append(self._group_by)
        self._vector_data = load_vector_data(
            vector_data, measurements=load_measurements
        )
        self._geobox = geobox_maker.from_vector(self._vector_data)
        self._grid_coords = affine_to_coords(
            self._geobox.affine, self._geobox.width, self._geobox.height
        )
        if self._geobox.crs is not None:
            self._vector_data = self._vector_data.to_crs(self._geobox.crs)

        self._measurements = measurements
        if self._measurements is None:
            self._measurements = self._vector_data.columns.tolist()
            self._measurements.remove("geometry")

        if categorical_enums is not None:
            for column_name, categories in categorical_enums.items():
                if column_name not in self._measurements:
                    continue
                category_type = pandas.api.types.CategoricalDtype(
                    categories=sorted(set(categories)) + ["nodata"]
                )
                self._vector_data[column_name] = self._vector_data[column_name].astype(
                    category_type
                )

        # get categorical enumerations if they exist
        self._categorical_enums: dict[str, list] = {
            categorical_column: self._vector_data[categorical_column].cat.categories
            for categorical_column in self._vector_data.select_dtypes(
                ["category"]
            ).columns
            if categorical_column in self._measurements
        }

        self._datetime_measurements: tuple[str, ...] = ()
        if datetime_measurements is not None:
            self._datetime_measurements = tuple(
                set(datetime_measurements) & set(self._measurements)
            )

        # convert to datetime
        for datetime_measurement in self._datetime_measurements:  # type: ignore
            date_data = pandas.to_datetime(self._vector_data[datetime_measurement])
            try:
                date_data = date_data.dt.tz_convert("UTC")
            except TypeError:
                pass
            self._vector_data[datetime_measurement] = date_data.dt.tz_localize(
                None
            ).astype("datetime64[ns]")

        if self._group_by:
            self._vector_data = self._vector_data.groupby(self._group_by)
            try:
                self._measurements.remove(self._group_by)
            except ValueError:
                pass

    def make_geocube(
        self,
        interpolate_na_method: Optional[Literal["linear", "nearest", "cubic"]] = None,
        rasterize_function: Optional[Callable[..., Optional[NDArray]]] = None,
    ) -> xarray.Dataset:
        """
        Rasterize vector data into an ``xarray`` object.  Each measurement will be a
        data variable in the :class:`xarray.Dataset`.

        See the `xarray documentation <http://xarray.pydata.org>`_
        for usage of the :class:`xarray.Dataset` and :class:`xarray.DataArray` objects.

        Parameters
        ----------
        interpolate_na_method:  {'linear', 'nearest', 'cubic'}, optional
            This is the method for interpolation to use to fill in the nodata with
            :func:`scipy.interpolate.griddata`.
        rasterize_function: function, optional
            Function to rasterize geometries. Other options are available in
            :mod:`geocube.rasterize`. Default is :func:`geocube.rasterize.rasterize_image`.

        Returns
        --------
        :class:`xarray.Dataset`:
            Requested data in a :class:`xarray.Dataset`.

        """
        self._rasterize_function = (
            rasterize_image if rasterize_function is None else rasterize_function  # type: ignore
        )
        return self._get_dataset(
            interpolate_na_method=interpolate_na_method,
        )

    @staticmethod
    def _get_attrs(
        measurement_name: str, fill_value: float
    ) -> dict[str, Union[str, float]]:
        """
        Get attributes for data array.

        Parameters
        ----------
        measurement_name: str
            The measurement name.
        fill_value: int or float
            The fill value.

        Returns
        -------
        dict: Dict with attributes for data array.
        """
        return {
            "name": measurement_name,
            "long_name": measurement_name,
            "_FillValue": fill_value,
        }

    def _update_time_attrs(self, attrs: dict[str, Any], image_data: NDArray) -> None:
        """
        Update attributes and nodata values for time grid.

        Parameters
        ----------
        attrs: dict
            Dict with attributes for data array.
        image_data: :obj:`numpy.ndarray`
            Array with image data.

        Returns
        -------
        None
        """
        attrs["units"] = "seconds from 1970-01-01T00:00:00"
        attrs["_FillValue"] = 0
        image_data[image_data == self._fill] = 0.0

    def _get_dataset(
        self,
        interpolate_na_method: Optional[str],
    ) -> xarray.Dataset:
        """
        Parameters
        ----------
        vector_data: :obj:`geopandas.GeoDataFrame`
            A GeoDataFrame containing the vector data.
        interpolate_na_method:  {'linear', 'nearest', 'cubic'}, optional
            This is the method for interpolation to use to fill in the nodata with
            :func:`scipy.interpolate.griddata`.

        Returns
        --------
        :obj:`xarray.Dataset`:
            Requested data in a :obj:`xarray.Dataset`.

        """
        data_vars = {}
        for measurement in self._measurements or ():
            if self._group_by:
                grid_array = self._get_grouped_grid(
                    self._vector_data[[measurement, "geometry"]],
                    measurement_name=measurement,
                )
            else:
                grid_array = self._get_grid(
                    self._vector_data[[measurement, "geometry"]],
                    measurement_name=measurement,
                )
            if grid_array is not None:
                data_vars[measurement] = grid_array

        if self._group_by:
            self._grid_coords[self._group_by] = list(self._vector_data.groups.keys())  # type: ignore

        out_xds = xarray.Dataset(data_vars=data_vars, coords=self._grid_coords)

        for categorical_measurement, categoral_enums in self._categorical_enums.items():
            enum_var_name = f"{categorical_measurement}_categories"
            cat_attrs = dict(out_xds[categorical_measurement].attrs)
            cat_attrs["categorical_mapping"] = enum_var_name
            out_xds[categorical_measurement].attrs = cat_attrs
            out_xds[enum_var_name] = categoral_enums

        out_xds.rio.write_transform(self._geobox.affine, inplace=True)
        out_xds.rio.write_crs(str(self._geobox.crs), inplace=True)
        out_xds.rio.write_coordinate_system(inplace=True)
        if interpolate_na_method is not None:
            return out_xds.rio.interpolate_na(method=interpolate_na_method)

        return out_xds

    def _get_grouped_grid(
        self,
        grouped_dataframe: geopandas.GeoDataFrame,
        measurement_name: str,
    ) -> Optional[tuple]:
        """Retrieve the variable data to append to the ssurgo :obj:`xarray.Dataset`.
        This method is designed specifically to work on a dataframe that has
        been grouped.

        Parameters
        ----------
        grouped_dataframe: pandas GroupBy object
            A pandas dataframe in as a GroupBy object.
        measurement_name: str
            Attributes name or list of names to be included. If a list is specified,
            the measurements will be returned in the order requested.
            By default all available measurements are included.

        Returns
        -------
        Optional[tuple]: Options needed to create an :obj:`xarray.DataArray`.

        """
        logger = get_logger()

        image_data = []
        df_group = None
        fill_value = self._fill
        for _, df_group in grouped_dataframe:
            fill_value = (
                self._fill
                if str(df_group[measurement_name].dtype) != "category"
                else -1
            )
            image = self._rasterize_function(
                geometry_array=df_group.geometry,
                data_values=_format_series_data(df_group[measurement_name]).values,
                geobox=self._geobox,
                grid_coords=self._grid_coords,
                fill=fill_value,
            )
            if image is None:
                logger.warning(
                    f"Skipping attribute {measurement_name} due to missing data..."
                )
                return None

            image_data.append(image)

        attrs = self._get_attrs(measurement_name, fill_value)
        image_data = numpy.array(image_data)
        # it was converted to numeric date value
        if df_group is not None and "datetime" in str(df_group[measurement_name].dtype):
            self._update_time_attrs(attrs, image_data)

        return (
            (self._group_by, "y", "x"),
            image_data,
            attrs,
            {"grid_mapping": DEFAULT_GRID_MAP},
        )

    def _get_grid(
        self, dataframe: geopandas.GeoDataFrame, measurement_name: str
    ) -> Optional[tuple]:
        """Retrieve the variable data to append to the ssurgo :obj:`xarray.Dataset`
        from a regular :obj:`geopandas.GeoDataFrame`.

        Parameters
        ----------
        dataframe: :obj:`geopandas.GeoDataFrame`
            A geopandas GeoDataFrame object to rasterize.
        measurement_name: str
            Attributes name or list of names to be included. If a list is specified,
            the measurements will be returned in the order requested.
            By default all available measurements are included.

        Returns
        -------
        Optional[tuple]: Options needed to create an :obj:`xarray.DataArray`.

        """
        logger = get_logger()
        fill_value = (
            self._fill if str(dataframe[measurement_name].dtype) != "category" else -1
        )
        image_data = self._rasterize_function(
            geometry_array=dataframe.geometry,
            data_values=_format_series_data(dataframe[measurement_name]).values,
            geobox=self._geobox,
            grid_coords=self._grid_coords,
            fill=fill_value,
        )
        if image_data is None:
            logger.warning(
                f"Skipping attribute {measurement_name} due to missing data..."
            )
            return None

        attrs = self._get_attrs(measurement_name, fill_value)

        # it was converted to numeric date value
        if "datetime" in str(dataframe[measurement_name].dtype):
            self._update_time_attrs(attrs, image_data)

        return (
            ("y", "x"),
            numpy.array(image_data),
            attrs,
            {"grid_mapping": DEFAULT_GRID_MAP},
        )



"""
This module is an extension for xarray to provide support for vector datasets.
"""
import geopandas
import numpy
import rioxarray  # noqa: F401 pylint: disable=unused-import
import xarray
from shapely.wkb import dumps, loads

dumps_v = numpy.vectorize(dumps)
loads_v = numpy.vectorize(loads)


def from_geodataframe(in_geodataframe):
    """
    Create an xarray object from a geodataframe.
    """
    # pylint: disable=protected-access
    geodf = in_geodataframe.copy().set_index("geometry")
    geox = xarray.Dataset.from_dataframe(geodf)
    # hack to get around dimension error when
    # writing CRS
    geox.rio._x_dim = "x"
    geox.rio._y_dim = "y"
    geox.rio.write_crs(in_geodataframe.crs, inplace=True)
    geox.rio._x_dim = None
    geox.rio._y_dim = None
    return geox


def open_dataset(*args, **kwargs):
    """
    Open the vxarray exported netCDF file as a Dataset.

    Returns
    -------
    :obj:`xarray.Dataset`

    """
    xds = xarray.open_dataset(*args, **kwargs)
    return xds.assign_coords(geometry=loads_v(xds.coords["geometry"].values))


class BaseVectorX:
    """This is the base for the vector GIS extension."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _validate_operation(self):
        if set(self._obj.dims) - {"geometry"}:
            raise ValueError(
                "Extra dimensions found. Please slice/squeeze data"
                " so that the only dimension is 'geometry'."
            )

    def to_geodataframe(self):
        """
        Convert to a GeoDataFrame.

        Returns
        -------
        :obj:`geopandas.GeoDataFrame`

        """
        self._validate_operation()
        out_obj = self._obj.drop_vars(self._obj.rio.grid_mapping)
        extra_coords = list(set(list(out_obj.coords)) - {"geometry"})
        if extra_coords:
            out_obj = out_obj.copy().reset_coords(extra_coords)
        geodf = geopandas.GeoDataFrame(out_obj.to_dataframe().reset_index())
        geodf.crs = self._obj.rio.crs
        return geodf

    def to_netcdf(self, *args, **kwargs):
        """
        Write the data to a netCDF file.
        """
        self._obj.assign_coords(
            geometry=dumps_v(self._obj.coords["geometry"].values)
        ).to_netcdf(*args, **kwargs)


@xarray.register_dataarray_accessor("vector")
class VectorArray(BaseVectorX):
    """This is the vector GIS extension for :class:`xarray.DataArray`"""

    def plot(self, *args, **kwargs):
        """
        Helper method to plot the geometry data via geopandas.

        Parameters
        ----------
        *args, **kwargs:
            Arguments to pass to the plot call.
        """
        self.to_geodataframe().plot(column=self._obj.name, *args, **kwargs)


@xarray.register_dataset_accessor("vector")
class VectorDataset(BaseVectorX):
    """This is the vector GIS extension for :class:`xarray.Dataset`"""

    def plot(self, column, *args, **kwargs):
        """
        Helper method to plot the geometry data via geopandas.

        Parameters
        ----------
        column: str
            The name of the column (variable) to plot.
        *args, **kwargs:
            Arguments to pass to the plot call.
        """
        self._obj[column].vector.plot(*args, **kwargs)



"""
Module for vector methods
"""
import geopandas
import numpy
import rasterio.features
import rioxarray  # noqa: F401 pylint: disable=unused-import
import shapely.geometry
import xarray


def vectorize(data_array: xarray.DataArray) -> geopandas.GeoDataFrame:
    """
    .. versionadded:: 0.4.0

    Powered by: :func:`rasterio.features.shapes`

    Convert 2D :class:`xarray.DataArray` into
    a :class:`geopandas.GeoDataFrame`.

    The ``nodata``,  ``CRS``, and ``transform`` of the :class:`xarray.DataArray`
    are determined using ``rioxarray``.

    Helpful references:

    - https://corteva.github.io/rioxarray/stable/getting_started/crs_management.html
    - https://corteva.github.io/rioxarray/stable/getting_started/nodata_management.html


    Parameters
    ----------
    data_array: xarray.DataArray
        Input 2D DataArray raster.

    Returns
    -------
    geopandas.GeoDataFrame
    """
    # nodata mask
    mask = None
    if data_array.rio.nodata is not None:
        if numpy.isnan(data_array.rio.nodata):
            mask = ~data_array.isnull()
        else:
            mask = data_array != data_array.rio.nodata

    # vectorize generator
    vectorized_data = (
        (value, shapely.geometry.shape(polygon))
        for polygon, value in rasterio.features.shapes(
            data_array,
            transform=data_array.rio.transform(),
            mask=mask,
        )
    )
    return geopandas.GeoDataFrame(
        vectorized_data,
        columns=[data_array.name, "geometry"],
        crs=data_array.rio.crs,
    )



"""Top-level package for geocube."""

__author__ = """Geocube Contributors"""
__email__ = "alansnow21@gmail.com"

import importlib.metadata

from geocube._show_versions import show_versions  # noqa

__version__ = importlib.metadata.version(__package__)
